@@ -12,20 +12,27 @@ from AnyQt.QtCore import (
 )
 from AnyQt.QtCore import pyqtSignal as Signal, pyqtSlot as Slot
 from AnyQt.QtGui import (
-    QPixmap, QPen, QBrush, QColor, QPainter, QImageReader
+    QPixmap, QPen, QBrush, QColor, QPainter, QPainterPath, QImageReader,
+    QFontMetrics
 )
 from AnyQt.QtWidgets import (
-    QGraphicsView, QGraphicsWidget, QGraphicsItem,
+    QGraphicsScene, QGraphicsView, QGraphicsWidget, QGraphicsItem,
+    QGraphicsRectItem,
     QGraphicsLinearLayout,
     QGraphicsGridLayout, QSizePolicy, QApplication, QStyle, QShortcut,
-    QFormLayout)
+    QFormLayout, QLabel)
 from Orange.widgets import widget, gui, settings
-from Orange.widgets.utils.itemmodels import VariableListModel
+from Orange.widgets.settings import ContextSetting
+from Orange.widgets.utils.annotated_data import (
+    create_annotated_table, create_groups_table)
+from Orange.widgets.utils.colorpalette import ColorPaletteGenerator
+from Orange.widgets.utils.itemmodels import VariableListModel, DomainModel
 from Orange.widgets.utils.overlay import proxydoc
 from Orange.widgets.widget import Input, Output, OWWidget, Msg
 
 from orangecontrib.imageanalytics.image_grid import ImageGrid
-from orangecontrib.imageanalytics.widgets.owimageviewer import ImageLoader, Preview, GraphicsScene
+from orangecontrib.imageanalytics.widgets.owimageviewer import (
+    ImageLoader, Preview)
 
 _log = logging.getLogger(__name__)
 
@@ -37,6 +44,9 @@ _ImageItem = namedtuple(
      "future"]  # Future instance yielding an QImage
 )
 
+DEFAULT_SELECTION_BRUSH = QBrush(QColor(217, 232, 252, 192))
+DEFAULT_SELECTION_PEN = QPen(QColor(125, 162, 206, 192))
+
 
 class OWImageGrid(widget.OWWidget):
     name = "Image Grid"
@@ -47,10 +57,12 @@ class OWImageGrid(widget.OWWidget):
     graph_name = "scene"
 
     class Inputs:
-        data = Input("Embeddings", Orange.data.Table)
+        data = Input("Embeddings", Orange.data.Table, default=True)
         data_subset = Input("Data Subset", Orange.data.Table)
 
     class Outputs:
+        selected_data = Output(
+            "Selected Images", Orange.data.Table, default=True)
         data = Output("Images", Orange.data.Table)
 
     settingsHandler = settings.DomainContextHandler()
@@ -61,6 +73,8 @@ class OWImageGrid(widget.OWWidget):
 
     imageAttr = settings.ContextSetting(0)
     imageSize = settings.Setting(100)
+    label_attr = settings.ContextSetting(None, required=ContextSetting.OPTIONAL)
+    label_selected = settings.Setting(True)
 
     auto_update = settings.Setting(True)
     auto_commit = settings.Setting(True)
@@ -76,13 +90,15 @@ class OWImageGrid(widget.OWWidget):
 
         self.data = None
         self.data_subset = None
+        self.subset_indices = []
+        self.nonempty = []
 
         self.allAttrs = []
         self.stringAttrs = []
         self.domainAttrs = []
+        self.label_model = DomainModel(placeholder="(No labels)")
 
-        self.selected_indices = []
-        self.subset_indices = []
+        self.selection = None
 
         #: List of _ImageItems
         self.items = []
@@ -94,14 +110,15 @@ class OWImageGrid(widget.OWWidget):
             self.controlArea, self, "imageAttr",
             box="Image Filename Attribute",
             tooltip="Attribute with image filenames",
-            callback=self.changeImageAttr,
+            callback=self.change_image_attr,
             contentsLength=12,
             addSpace=True,
         )
 
         # cell fit (resize or crop)
-        self.cellFitRB = gui.radioButtons(self.controlArea, self, "cell_fit", ["Resize", "Crop"],
-                                          box="Image cell fit", callback=self.setCrop)
+        self.cellFitRB = gui.radioButtons(
+            self.controlArea, self, "cell_fit", ["Resize", "Crop"],
+            box="Image cell fit", callback=self.set_crop)
 
         self.gridSizeBox = gui.vBox(self.controlArea, "Grid size")
 
@@ -112,8 +129,12 @@ class OWImageGrid(widget.OWWidget):
             verticalSpacing=10
         )
 
-        self.colSpinner = gui.spin(self.gridSizeBox, self, "columns", minv=2, maxv=40, callback=self.updateSize)
-        self.rowSpinner = gui.spin(self.gridSizeBox, self, "rows", minv=2, maxv=40, callback=self.updateSize)
+        self.colSpinner = gui.spin(
+            self.gridSizeBox, self, "columns", minv=2, maxv=40,
+            callback=self.update_size)
+        self.rowSpinner = gui.spin(
+            self.gridSizeBox, self, "rows", minv=2, maxv=40,
+            callback=self.update_size)
 
         form.addRow("Columns:", self.colSpinner)
         form.addRow("Rows:", self.rowSpinner)
@@ -121,15 +142,27 @@ class OWImageGrid(widget.OWWidget):
         gui.separator(self.gridSizeBox, 10)
         self.gridSizeBox.layout().addLayout(form)
 
-        gui.button(self.gridSizeBox, self, "Set size automatically", callback=self.autoSetSize)
+        gui.button(
+            self.gridSizeBox, self, "Set size automatically",
+            callback=self.auto_set_size)
+
+        self.label_box = gui.vBox(self.controlArea, "Labels")
+
+        # labels control
+        self.label_attr_cb = gui.comboBox(
+            self.label_box, self, "label_attr",
+            tooltip="Show labels",
+            callback=self.update_size,
+            addSpace=True,
+            model=self.label_model
+        )
 
         gui.rubber(self.controlArea)
 
         # auto commit
-        self.autoCommitBox = gui.auto_commit(self.controlArea, self, "auto_commit", "Apply",
-                                             checkbox_label="Apply automatically")
-
-        self.info = gui.widgetLabel(gui.vBox(self.controlArea, "Info"), "Waiting for input.\n")
+        self.autoCommitBox = gui.auto_commit(
+            self.controlArea, self, "auto_commit", "Apply",
+            checkbox_label="Apply automatically")
 
         self.image_grid = None
         self.cell_fit = 0
@@ -142,7 +175,7 @@ class OWImageGrid(widget.OWWidget):
         )
         self.mainArea.layout().addWidget(self.thumbnailView)
         self.scene = self.thumbnailView.scene()
-        self.scene.selectionChanged.connect(self.onSelectionChanged)
+        self.scene.selectionChanged.connect(self.on_selection_changed)
         self.loader = ImageLoader(self)
 
     def process(self, size_x=0, size_y=0):
@@ -152,9 +185,9 @@ class OWImageGrid(widget.OWWidget):
     def sizeHint(self):
         return QSize(600, 600)
 
-    # checks the input data for the right meta-attributes and finds the image filename.
+    # checks the input data for the right meta-attributes and finds images
     @Inputs.data
-    def setData(self, data):
+    def set_data(self, data):
         self.closeContext()
         self.clear()
         self.Warning.no_valid_data.clear()
@@ -180,20 +213,21 @@ class OWImageGrid(widget.OWWidget):
 
             self.imageAttrCB.setModel(VariableListModel(self.stringAttrs))
 
+            # set label combo labels
+            self.label_model.set_domain(domain)
             self.openContext(data)
+            self.label_attr = self.label_model[0]
+            self.imageAttr = max(
+                min(self.imageAttr, len(self.stringAttrs) - 1), 0)
 
-            self.imageAttr = max(min(self.imageAttr, len(self.stringAttrs) - 1), 0)
-
-            if self.isValidData():
+            if self.is_valid_data():
                 self.image_grid = ImageGrid(data)
-                self.setupScene()
+                self.setup_scene()
             else:
                 self.Warning.no_valid_data()
-        else:
-            self.info.setText("Waiting for input.\n")
 
     @Inputs.data_subset
-    def setDataSubset(self, data_subset):
+    def set_data_subset(self, data_subset):
         self.data_subset = data_subset
 
     def clear(self):
@@ -201,13 +235,15 @@ class OWImageGrid(widget.OWWidget):
         self.image_grid = None
         self.error()
         self.imageAttrCB.clear()
-        self.clearScene()
+        self.label_attr_cb.clear()
+        self.clear_scene()
 
-    def isValidData(self):
+    def is_valid_data(self):
         return self.data and self.stringAttrs and self.domainAttrs
 
     # loads the images and places them into the viewing area
-    def setupScene(self, process_grid=True):
+    def setup_scene(self, process_grid=True):
+        self.clear_scene()
         self.error()
         if self.data:
             attr = self.stringAttrs[self.imageAttr]
@@ -223,16 +259,26 @@ class OWImageGrid(widget.OWWidget):
             self.thumbnailView.setFixedRowCount(self.rows)
 
             for i, inst in enumerate(self.image_grid.image_list):
-                thumbnail = GraphicsThumbnailWidget(QPixmap(), crop=self.cell_fit == 1)
+                label_text = (str(inst[self.label_attr])
+                    if self.label_attr is not None else "")
+                if label_text == "?":
+                    label_text = ""
+
+                thumbnail = GraphicsThumbnailWidget(
+                    QPixmap(), crop=self.cell_fit == 1,
+                    add_label=self.label_selected and
+                    self.label_attr is not None, text=label_text)
                 thumbnail.setThumbnailSize(size)
                 thumbnail.instance = inst
                 self.thumbnailView.addThumbnail(thumbnail)
 
-                if not np.isfinite(inst[attr]) or inst[attr] == "?":  # skip missing
+                if not np.isfinite(inst[attr]) or inst[attr] == "?":
+                    # skip missing
                     future, url = None, None
                 else:
-                    url = self.urlFromValue(inst[attr])
+                    url = self.url_from_value(inst[attr])
                     thumbnail.setToolTip(url.toString())
+                    self.nonempty.append(i)
 
                     if url.isValid() and url.isLocalFile():
                         reader = QImageReader(url.toLocalFile())
@@ -271,34 +317,43 @@ class OWImageGrid(widget.OWWidget):
 
                             thumb.setPixmap(pixmap)
 
-                            self._noteCompleted(future)
+                            self._note_completed(future)
                     else:
                         future = None
 
                 self.items.append(_ImageItem(i, thumbnail, url, future))
 
-            if any(not it.future.done() if it.future else False for it in self.items):
-                self.info.setText("Retrieving...\n")
-            else:
-                self._updateStatus()
-                self.applySubset()
+            if not any(
+                    not it.future.done() if it.future
+                    else False for it in self.items):
+                self._update_status()
+                self.apply_subset()
+                self.update_selection()
 
     def handleNewSignals(self):
         self.Warning.incompatible_subset.clear()
         self.subset_indices = []
 
-        if self.data and self.data_subset:
+        if self.data and self.data_subset and len(self.stringAttrs) > 0:
             transformed = self.data_subset.transform(self.data.domain)
-            if np.all(self.data.domain.metas == self.data_subset.domain.metas):
-                indices = {e.id for e in transformed}
-                self.subset_indices = [ex.id in indices for ex in self.data]
+            attr = self.stringAttrs[self.imageAttr]
 
+            # for the subset we need to check if it contains the image
+            # attribute, if all indices from the subset are in the original
+            # array, and if same image on the same index
+            data_indices = [e.id for e in self.data]
+            if attr in self.data_subset.domain \
+                and all(row.id in data_indices and self.data[np.where(
+                    data_indices == row.id)[0][0], attr] == row[attr]
+                        for row in transformed):
+                indices = {e.id for e in transformed}
+
+                self.subset_indices = [ex.id in indices for ex in self.data]
             else:
                 self.Warning.incompatible_subset()
+        self.apply_subset()
 
-        self.applySubset()
-
-    def urlFromValue(self, value):
+    def url_from_value(self, value):
         base = value.variable.attributes.get("origin", "")
         if QDir(base).exists():
             base = QUrl.fromLocalFile(base)
@@ -312,7 +367,7 @@ class OWImageGrid(widget.OWWidget):
         url = base.resolved(QUrl(str(value)))
         return url
 
-    def _cancelAllFutures(self):
+    def cancel_all_futures(self):
         for item in self.items:
             if item.future is not None:
                 item.future.cancel()
@@ -321,22 +376,24 @@ class OWImageGrid(widget.OWWidget):
                     item.future._reply.deleteLater()
                     item.future._reply = None
 
-    def clearScene(self):
-        self._cancelAllFutures()
+    def clear_scene(self):
+        self.cancel_all_futures()
         self.items = []
+        self.nonempty = []
+        self.selection = None
         self.thumbnailView.clear()
         self._errcount = 0
         self._successcount = 0
 
-    def changeImageAttr(self):
-        self.clearScene()
-        if self.isValidData():
-            self.setupScene()
+    def change_image_attr(self):
+        self.clear_scene()
+        if self.is_valid_data():
+            self.setup_scene()
 
-    def thumbnailItems(self):
+    def thumbnail_items(self):
         return [item.widget for item in self.items]
 
-    def updateSize(self):
+    def update_size(self):
         try:
             self.process(self.columns, self.rows)
             self.colSpinner.setMinimum(2)
@@ -350,48 +407,106 @@ class OWImageGrid(widget.OWWidget):
             self.rowSpinner.setMinimum(self.rows)
             return
 
-        self.clearScene()
-        if self.isValidData():
-            self.setupScene(process_grid=False)
+        self.clear_scene()
+        if self.is_valid_data():
+            self.setup_scene(process_grid=False)
 
-    def setCrop(self):
+    def set_crop(self):
         self.thumbnailView.setCrop(self.cell_fit == 1)
 
-    def autoSetSize(self):
-        self.clearScene()
-        if self.isValidData():
-            self.setupScene()
+    def auto_set_size(self):
+        self.clear_scene()
+        if self.is_valid_data():
+            self.setup_scene()
 
-    def applySubset(self):
+    def apply_subset(self):
         if self.image_grid:
-            subset_indices = self.subset_indices if self.subset_indices else [True] * len(self.items)
-            ordered_subset_indices = self.image_grid.order_to_grid(subset_indices)
+            subset_indices = (self.subset_indices if self.subset_indices
+                else [True] * len(self.items))
+            ordered_subset_indices = self.image_grid.order_to_grid(
+                subset_indices)
 
             for item, in_subset in zip(self.items, ordered_subset_indices):
                 item.widget.setSubset(in_subset)
 
-    def onSelectionChanged(self):
-        selected = [item for item in self.items if item.widget.isSelected() and item.url != ""]
-        self.selected_indices = [item.index for item in selected]
+    def on_selection_changed(self, selected_items, keys):
+        if self.selection is None:
+            self.selection = np.zeros(len(self.items), dtype=np.uint8)
+
+        # newly selected
+        indices = [item.index for item in self.items
+                   if item.widget in selected_items]
+
+        # Remove from selection
+        if keys & Qt.AltModifier:
+            self.selection[indices] = 0
+        # Append to the last group
+        elif keys & Qt.ShiftModifier and keys & Qt.ControlModifier:
+            self.selection[indices] = np.max(self.selection)
+        # Create a new group
+        elif keys & Qt.ShiftModifier:
+            self.selection[indices] = np.max(self.selection) + 1
+        # No modifiers: new selection
+        else:
+            self.selection = np.zeros(len(self.items), dtype=np.uint8)
+            self.selection[indices] = 1
+
+        self.update_selection()
         self.commit()
 
     def commit(self):
         if self.data:
-            if self.selected_indices:
-                selected = self.image_grid.image_list[self.selected_indices]
+            # add Group column (group number)
+            self.Outputs.selected_data.send(
+                create_groups_table(self.image_grid.image_list, self.selection,
+                                    False, "Group"))
+
+            # filter out empty cells - keep indices of cells that contain images
+            # add Selected column
+            # (Yes/No if one group, else Unselected or group number)
+            if self.selection is not None and np.max(self.selection) > 1:
+                out_data = create_groups_table(
+                    self.image_grid.image_list[self.nonempty],
+                    self.selection[self.nonempty])
             else:
-                selected = None
-            self.Outputs.data.send(selected)
+                out_data = create_annotated_table(
+                    self.image_grid.image_list[self.nonempty],
+                    np.nonzero(self.selection[self.nonempty]))
+            self.Outputs.data.send(out_data)
+
         else:
             self.Outputs.data.send(None)
+            self.Outputs.selected_data.send(None)
+
+    def update_selection(self):
+        if self.selection is not None:
+            pen, brush = self.compute_colors()
+
+            for s, item, p, b in zip(self.selection, self.items, pen, brush):
+                item.widget.setSelected(s > 0)
+                item.widget.setSelectionColor(p, b)
+
+    # Adapted from Scatter Plot Graph (change brush instead of pen)
+    def compute_colors(self):
+        no_brush = DEFAULT_SELECTION_BRUSH
+        sels = np.max(self.selection)
+        if sels == 1:
+            brushes = [no_brush, no_brush]
+        else:
+            palette = ColorPaletteGenerator(number_of_colors=sels + 1)
+            brushes = [no_brush] + [QBrush(palette[i]) for i in range(sels)]
+        brush = [brushes[a] for a in self.selection]
+
+        pen = [DEFAULT_SELECTION_PEN] * len(self.items)
+        return pen, brush
 
     def send_report(self):
-        if self.isValidData():
+        if self.is_valid_data():
             items = [("Number of images", len(self.data))]
             self.report_items(items)
             self.report_plot("Grid", self.scene)
 
-    def _noteCompleted(self, future):
+    def _note_completed(self, future):
         # Note the completed future's state
         if future.cancelled():
             return
@@ -402,32 +517,19 @@ class OWImageGrid(widget.OWWidget):
         else:
             self._successcount += 1
 
-        self._updateStatus()
+        self._update_status()
 
-    def _updateStatus(self):
+    def _update_status(self):
         count = len([item for item in self.items if item.future is not None])
-        self.info.setText(
-            "Retrieving:\n" +
-            "{} of {} images".format(self._successcount, count))
 
         if self._errcount + self._successcount == count:
-            if self._errcount:
-                self.info.setText(
-                    "Done:\n" +
-                    "{} images, {} errors".format(count, self._errcount)
-                )
-            else:
-                self.info.setText(
-                    "Done:\n" +
-                    "{} images".format(count)
-                )
             attr = self.stringAttrs[self.imageAttr]
             if self._errcount == count and "type" not in attr.attributes:
                 self.error("No images found! Make sure the '%s' attribute "
                            "is tagged with 'type=image'" % attr.name)
 
     def onDeleteWidget(self):
-        self._cancelAllFutures()
+        self.cancel_all_futures()
         self.clear()
 
 
@@ -442,6 +544,7 @@ Changes:
 - layout is fixed instead of autoreflowing
 - resizing policy for individual Pixmap widgets is now fixed
 instead of auto-stretch
+- improved selection (with groups)
 """
 
 
@@ -504,7 +607,8 @@ class GraphicsPixmapWidget(QGraphicsWidget):
 
         rect = self.contentsRect()
         pixsize = QSizeF(self._pixmap.size())
-        aspectmode = (Qt.KeepAspectRatio if self._keepAspect else Qt.IgnoreAspectRatio)
+        aspectmode = (
+            Qt.KeepAspectRatio if self._keepAspect else Qt.IgnoreAspectRatio)
 
         if self._crop:
             height, width = pixsize.height(), pixsize.width()
@@ -528,7 +632,7 @@ class GraphicsPixmapWidget(QGraphicsWidget):
         if self._subset:
             painter.setOpacity(1.0)
         else:
-            painter.setOpacity(0.5)
+            painter.setOpacity(0.35)
 
         pixrect.moveCenter(rect.center())
         painter.save()
@@ -540,8 +644,19 @@ class GraphicsPixmapWidget(QGraphicsWidget):
         painter.restore()
 
 
+class ElidedLabel(QLabel):
+    def paintEvent( self, event ):
+        painter = QPainter(self)
+
+        metrics = QFontMetrics(self.font())
+        elided = metrics.elidedText(self.text(), Qt.ElideRight, self.width())
+
+        painter.drawText(self.rect(), self.alignment(), elided)
+
+
 class GraphicsThumbnailWidget(QGraphicsWidget):
-    def __init__(self, pixmap, parentItem=None, crop=False, in_subset=True, **kwargs):
+    def __init__(self, pixmap, parentItem=None, crop=False, in_subset=True,
+                 add_label=False, text="", **kwargs):
         super().__init__(parentItem, **kwargs)
         self.setFocusPolicy(Qt.StrongFocus)
         self._size = QSizeF()
@@ -554,10 +669,24 @@ class GraphicsThumbnailWidget(QGraphicsWidget):
         self.pixmapWidget = GraphicsPixmapWidget(pixmap, self)
         self.pixmapWidget.setCrop(crop)
         self.pixmapWidget.setSubset(in_subset)
+        self.selectionBrush = DEFAULT_SELECTION_BRUSH
+        self.selectionPen = DEFAULT_SELECTION_PEN
 
         layout.addItem(self.pixmapWidget)
-        layout.setAlignment(self.pixmapWidget, Qt.AlignCenter)
 
+        self.label = None
+        if add_label:
+            l1 = ElidedLabel(text)
+            l1.setStyleSheet("background-color: rgba(255, 255, 255, 10);")
+            l1.setAlignment(Qt.AlignCenter)
+            l1.setFixedHeight(16)
+
+            self.label = l1
+            gs = QGraphicsScene()
+            w = gs.addWidget(l1)
+            layout.addItem(w)
+
+        layout.setAlignment(self.pixmapWidget, Qt.AlignCenter)
         self.setLayout(layout)
 
         self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
@@ -577,12 +706,19 @@ class GraphicsThumbnailWidget(QGraphicsWidget):
     def setSubset(self, in_subset):
         self.pixmapWidget.setSubset(in_subset)
 
+    def setSelectionColor(self, pen, brush):
+        self.selectionPen = pen
+        self.selectionBrush = brush
+        self.update()
+
     def pixmap(self):
         return self.pixmapWidget.pixmap()
 
     def setThumbnailSize(self, size):
         if self._size != size:
             self._size = QSizeF(size)
+            if self.label is not None:
+                self.label.setFixedWidth(size.width())
             self._updatePixmapSize()
 
     def paint(self, painter, option, widget=0):
@@ -593,9 +729,9 @@ class GraphicsThumbnailWidget(QGraphicsWidget):
             if option.state & QStyle.State_HasFocus:
                 painter.setPen(QPen(QColor(125, 0, 0, 192)))
             else:
-                painter.setPen(QPen(QColor(125, 162, 206, 192)))
+                painter.setPen(self.selectionPen)
             if option.state & QStyle.State_Selected:
-                painter.setBrush(QBrush(QColor(217, 232, 252, 192)))
+                painter.setBrush(self.selectionBrush)
             painter.drawRoundedRect(
                 QRectF(contents.topLeft(), self.geometry().size()), 3, 3)
             painter.restore()
@@ -637,10 +773,15 @@ class GraphicsThumbnailGrid(QGraphicsWidget):
         self.__layout.setSpacing(0)
         self.setLayout(self.__layout)
 
+    @property
+    def thumbnails(self):
+        return self.__thumbnails
+
     def resizeEvent(self, event):
         super().resizeEvent(event)
 
-        if event.newSize().width() != event.oldSize().width() or event.newSize().height() != event.oldSize().height():
+        if (event.newSize().width() != event.oldSize().width() or
+                event.newSize().height() != event.oldSize().height()):
             self.__gridlayout()
 
     def setGeometry(self, rect):
@@ -790,8 +931,8 @@ class GraphicsThumbnailGrid(QGraphicsWidget):
     def __scheduleLayout(self):
         if not self.__reflowPending:
             self.__reflowPending = True
-            QApplication.postEvent(self, QEvent(QEvent.LayoutRequest),
-                                   Qt.HighEventPriority)
+            QApplication.postEvent(
+                self, QEvent(QEvent.LayoutRequest), Qt.HighEventPriority)
 
     def event(self, event):
         if event.type() == QEvent.LayoutRequest:
@@ -854,15 +995,18 @@ class GraphicsThumbnailGrid(QGraphicsWidget):
     def __gridlayout(self):
         assert self.__layoutMode == GraphicsThumbnailGrid.FixedColumnCount
 
-        width = (self.size().width() - self.__columnCount * 10) / self.__columnCount
+        width = ((self.size().width() - self.__columnCount * 10) /
+                self.__columnCount)
         height = (self.size().height() - self.__rowCount * 10) / self.__rowCount
-        item_size = min(width, height)
 
         for item in self.__thumbnails:
+            label_size = item.label.height() + 1 if item.label is not None else 0
+            item_size = min(width, height - label_size)
             item.setThumbnailSize(QSizeF(item_size, item_size))
 
         self.__relayoutGrid(self.__columnCount)
 
+
     def __relayoutGrid(self, columnCount):
         layout = self.__layout
         if columnCount == layout.columnCount():
@@ -969,13 +1113,6 @@ class GraphicsThumbnailGrid(QGraphicsWidget):
         assert newcurrent is self.__thumbnails[index]
 
         if newcurrent is not None:
-            if not modifiers & (Qt.ShiftModifier | Qt.ControlModifier):
-                for item in self.__thumbnails:
-                    if item is not newcurrent:
-                        item.setSelected(False)
-                        # self.scene().clearSelection()
-
-            newcurrent.setSelected(True)
             newcurrent.setFocus(Qt.TabFocusReason)
             newcurrent.ensureVisible()
 
@@ -1021,6 +1158,10 @@ class ThumbnailView(QGraphicsView):
     def addThumbnail(self, thumbnail):
         self.__grid.addThumbnail(thumbnail)
 
+    @property
+    def grid(self):
+        return self.__grid
+
     @proxydoc(GraphicsThumbnailGrid.insertThumbnail)
     def insertThumbnail(self, index, thumbnail):
         self.__grid.insertThumbnail(index, thumbnail)
@@ -1083,7 +1224,8 @@ class ThumbnailView(QGraphicsView):
         return super().keyPressEvent(event)
 
     def __previewToggle(self):
-        if self.__previewWidget is None and self.__grid.currentItem() is not None:
+        if (self.__previewWidget is None and
+                self.__grid.currentItem() is not None):
             focusitem = self.__grid.currentItem()
             preview = self.__getPreviewWidget()
             preview.show()
@@ -1148,40 +1290,75 @@ class ThumbnailView(QGraphicsView):
         self.ensureVisible(QRectF(point, QSizeF(1, 1)), 5, 5),
 
 
-# TODO only loads when workdir is image dir (ImportImages keeps only the filename in the path)
-def main(argv=None):
-    import sys
-    from orangecontrib.imageanalytics.import_images import ImportImages
-    from orangecontrib.imageanalytics.image_embedder import ImageEmbedder
+class GraphicsScene(QGraphicsScene):
+    selectionRectPointChanged = Signal(QPointF)
 
-    if argv is None:
-        argv = sys.argv
+    # override the default signal since it should only be emitted when a
+    # selection is finished
+    selectionChanged = Signal(set, int)
 
-    argv = list(argv)
-    app = QApplication(argv)
+    def __init__(self, *args):
+        QGraphicsScene.__init__(self, *args)
+        self.selectionRect = None
 
-    if len(argv) > 1:
-        image_dir = argv[1]
-    else:
-        raise ValueError("Provide the image directory as the first argument.")
+    # TODO figure out how to keep items highlighted and prevent redrawing
+    #  during selection without disabling this method
+    def mousePressEvent(self, event):
+        QGraphicsScene.mousePressEvent(self, event)
 
-    import_images = ImportImages()
-    images, err = import_images(image_dir)
+    def mouseMoveEvent(self, event):
+        if event.buttons() & Qt.LeftButton:
+            screenPos = event.screenPos()
+            buttonDown = event.buttonDownScreenPos(Qt.LeftButton)
+            if (screenPos - buttonDown).manhattanLength() > 2.0:
+                self.updateSelectionRect(event)
+        QGraphicsScene.mouseMoveEvent(self, event)
 
-    image_embedder = ImageEmbedder()
-    embeddings, _, _ = image_embedder(images)
+    def mouseReleaseEvent(self, event):
+        QGraphicsScene.mouseReleaseEvent(self, event)
 
-    ow = OWImageGrid()
-    ow.show()
-    ow.raise_()
-    ow.setData(Orange.data.Table(embeddings))
-    rval = app.exec()
+        if event.button() == Qt.LeftButton:
+            modifiers = event.modifiers()
+            path = QPainterPath()
 
-    ow.saveSettings()
-    ow.onDeleteWidget()
+            # the mouse was moved
+            if self.selectionRect:
+                path.addRect(self.selectionRect.rect())
+                self.removeItem(self.selectionRect)
+                self.selectionRect = None
 
-    return rval
+            # the mouse was only clicked - create a selection area of 1x1 size
+            else:
+                rect = QRectF(
+                    event.buttonDownScenePos(Qt.LeftButton),
+                    QSizeF(1., 1.)).intersected(self.sceneRect())
+                path.addRect(rect)
+
+            self.setSelectionArea(path)
+            self.selectionChanged.emit(set(self.selectedItems()), modifiers)
+
+    def updateSelectionRect(self, event):
+        pos = event.scenePos()
+        buttonDownPos = event.buttonDownScenePos(Qt.LeftButton)
+        rect = QRectF(pos, buttonDownPos).normalized()
+        rect = rect.intersected(self.sceneRect())
+        if not self.selectionRect:
+            self.selectionRect = QGraphicsRectItem()
+            self.selectionRect.setBrush(QColor(10, 10, 10, 20))
+            self.selectionRect.setPen(QPen(QColor(200, 200, 200, 200)))
+            self.addItem(self.selectionRect)
+        self.selectionRect.setRect(rect)
+        self.selectionRectPointChanged.emit(pos)
 
 
 if __name__ == "__main__":
-    main()
+    from Orange.data import Table
+
+    from orangecontrib.imageanalytics.image_embedder import ImageEmbedder
+    from orangewidget.utils.widgetpreview import WidgetPreview
+
+    embedding, _, _ = ImageEmbedder(model="squeezenet")(
+        Table("https://datasets.biolab.si/core/bone-healing.xlsx"),
+        col="Image")
+
+    WidgetPreview(OWImageGrid).run(embedding)
